"""Middleware used by the history links service."""

from django.shortcuts import redirect
from django.utils.cache import add_never_cache_headers

from historylinks.registration import history_link_context_manager, default_history_link_manager


HISTORYLINK_MIDDLEWARE_FLAG = "historylinks.history_link_fallback_middleware_active"


class HistoryLinkFallbackMiddleware(object):

    """Middleware that attempts to rescue 404 responses with a redirect to it's new location."""

    def process_request(self, request):
        """Starts a new history link context."""
        request.META[(HISTORYLINK_MIDDLEWARE_FLAG, self)] = True
        history_link_context_manager.start()

    def _close_history_link_context(self, request):
        """Closes the history link context."""
        if request.META.get((HISTORYLINK_MIDDLEWARE_FLAG, self), False):
            del request.META[(HISTORYLINK_MIDDLEWARE_FLAG, self)]
            history_link_context_manager.end()

    def process_response(self, request, response):
        """Attempts to rescue 404 responses and closes the history link context."""
        # Close the history link context.
        self._close_history_link_context(request)
        # Attempt to rescue a 404 error.
        if response.status_code == 404:
            redirect_url = default_history_link_manager.get_current_url(request.path)
            if redirect_url and redirect_url != request.path:
                response = redirect(redirect_url, permanent=True)
                add_never_cache_headers(response)
                return response
        # Return the original response.
        return response

    def process_exception(self, request, exception):
        """Closes the history link context."""
        history_link_context_manager.invalidate()
        self._close_history_link_context(request)



from django.db import models
from django.test import TestCase

import historylinks
from historylinks.registration import RegistrationError


class HistoryLinkTestModel(models.Model):

    slug = models.SlugField(
        unique=True,
    )

    def get_absolute_url(self):
        return u"/{slug}/".format(slug=self.slug)

    class Meta:
        app_label = "auth"  # Hack: Cannot use an app_label that is under South control, due to http://south.aeracode.org/ticket/520


class RegistrationTest(TestCase):

    def testRegistration(self):
        # Register the model and test.
        historylinks.register(HistoryLinkTestModel)
        self.assertTrue(historylinks.is_registered(HistoryLinkTestModel))
        self.assertRaises(RegistrationError, lambda: historylinks.register(HistoryLinkTestModel))
        self.assertTrue(HistoryLinkTestModel in historylinks.get_registered_models())
        self.assertTrue(isinstance(historylinks.get_adapter(HistoryLinkTestModel), historylinks.HistoryLinkAdapter))
        # Unregister the model and text.
        historylinks.unregister(HistoryLinkTestModel)
        self.assertFalse(historylinks.is_registered(HistoryLinkTestModel))
        self.assertRaises(RegistrationError, lambda: historylinks.unregister(HistoryLinkTestModel))
        self.assertTrue(HistoryLinkTestModel not in historylinks.get_registered_models())
        self.assertRaises(RegistrationError, lambda: isinstance(historylinks.get_adapter(HistoryLinkTestModel)))


class HistoryLinkRedirectTest(TestCase):

    def setUp(self):
        historylinks.register(HistoryLinkTestModel)
        self.obj = HistoryLinkTestModel.objects.create(slug="foo")
        self.obj.slug = "bar"
        self.obj.save()

    def testRedirectsToNewURL(self):
        # Try a redirect.
        response = self.client.get("/foo/")
        self.assertEqual(response.status_code, 301)
        self.assertEqual(response["Location"], "http://testserver/bar/")
        # Try a 404.
        response = self.client.get("/baz/")
        self.assertEqual(response.status_code, 404)

    def tearDown(self):
        historylinks.unregister(HistoryLinkTestModel)



from django.core.management.base import NoArgsCommand
from django.db import transaction

from historylinks.registration import default_history_link_manager, _bulk_save_history_links


class Command(NoArgsCommand):

    help = "Builds the history links for all registered models."

    @transaction.commit_on_success
    def handle_noargs(self, **options):
        verbosity = int(options.get("verbosity", 1))
        link_count = 0
        # Create links.
        links_to_create = []
        for model in default_history_link_manager.get_registered_models():
            local_link_count = 0
            for obj in model._default_manager.all().iterator():
                local_link_count += 1
                links_to_create.extend(default_history_link_manager._update_obj_history_links_iter(obj))
                if verbosity == 3:
                    self.stdout.write("Refreshed history link for {obj}.\n".format(
                        obj=obj,
                    ))
            if verbosity == 2:
                self.stdout.write("Refreshed {local_link_count} history link(s) for {model}.\n".format(
                    local_link_count=local_link_count,
                    model=model._meta.verbose_name,
                ))
            link_count += local_link_count
        _bulk_save_history_links(links_to_create)
        if verbosity == 1:
            self.stdout.write("Refreshed {link_count} history links.\n".format(
                link_count=link_count,
            ))
